local _G = getfenv(0);
local LibStub = _G["LibStub"];
local AddOn = LibStub("AceAddon-3.0"):GetAddon("EquivalencePoints");

local AceSerializer = LibStub("AceSerializer-3.0");
local L = LibStub("AceLocale-3.0"):GetLocale("EquivalencePoints")

local AceLibrary = _G["AceLibrary"]; -- Get Ace2
local ItemBonusLib = AceLibrary("ItemBonusLib-1.0"); -- Ace2 atm
local TipHooker = AceLibrary("TipHooker-1.0"); -- Ace2 atm

local print = function(...) AddOn:PrintFunction(...); end
local debug = function(...) AddOn:DebugFunction(...); end

--[[ Locals ]]
local pcall, unpack = pcall, unpack;
local type, tostring, tonumber = type, tostring, tonumber;
local VersionCompare = AddOn.VersionCompare;
local pairs, ipairs = pairs, ipairs;
local rep, format, upper, find, gsub, strjoin, strsplit = string.rep, string.format, string.upper, string.find, string.gsub, strjoin, strsplit;
local insert, copy = table.insert, AddOn.table_copy;
local floor, round = math.floor, AddOn.math_round;
local rgb2hex = AddOn.rgb2hex;
local table, string, math = nil, nil, nil; -- Helpers

--[[ Begin AddOn ]]

function AddOn:OnInitialize()
	self:RegisterDB();
	self:Options();
	--self:ScheduleTimer("TipHookerHack", 1);
	self:SetCleanup();
	_, self.ClassEN = UnitClass("player");
	_, self.RaceEN = UnitRace("player"); -- NightElf, Dwarf, Human, Gnome, Draenei, Orc, Tauren, Troll, Scourge, BloodElf
	
	self.colors.class = {
		R = RAID_CLASS_COLORS[self.ClassEN]["r"],
		G = RAID_CLASS_COLORS[self.ClassEN]["g"],
		B = RAID_CLASS_COLORS[self.ClassEN]["b"],
	};
	
	if (not self.db.profile.AutoImportedEnhancer and self.ClassEN == "SHAMAN" and _G["Enhancer"]) then
		local import = true;
		for _,v in pairs(self.db.profile.Values) do
			if (v ~= 0) then
				import = false;
				break;
			end
		end
		
		if (import) then
			self.db.profile.AutoImportedEnhancer = true;
			self:ImportEnhancerAEP();
			print(L["Auto imported your |cffff0000AEP|r values from Enhancer!"]);
		end
	end
end

function AddOn:OnEnable()
	TipHooker:Hook(self.ProcessTooltip, "item");
	self:RegisterComm("EQP_VALUES_SENDER", "RecieveValues");
	self:News()
	
	AddOn.ClientVersion = (GetBuildInfo());
end

function AddOn:OnDisable()
	self:UnregisterAllEvents();
	self:UnregisterComm("EQP_SENDER");
	
	TipHooker:Unhook(self.ProcessTooltip, "item");
end

function AddOn:SetCleanup()
	local SetsDelete = {};
	
	for SetName in pairs(self.db.profile.HiddenSet) do
		if (not self.db.profile.ValueSet[SetName]) then
			insert(SetsDelete, SetName);
		end
	end
	
	for _, SetName in ipairs(SetsDelete) do
		self.db.profile.HiddenSet[SetName] = nil;
	end
end

function AddOn:TipHookerHack()
	local FunctionThatNeverExecutes = TipHooker.OnEventFrame:GetScript("OnEvent");
	pcall( FunctionThatNeverExecutes );
end

local ttS = rep(" ", 3); --tooltipSpacer was so fecking long to type each time ;) also it's nice if I want to change the number of indent spaces
function AddOn.ProcessTooltip(tooltip, name, link)
	debug("ProcessTooltip Start -----------------");
	if (link) then
		local self = AddOn;
		
		local sItemLink, ItemID, ItemRarity, _, ItemSubType = self:StripInfo(link);
		ItemRarity = (self.db.profile.MatchRarity and ItemRarity) or self.db.profile.MaxRarity;
		ItemRarity = (ItemRarity > self.db.profile.MaxRarity) and self.db.profile.MaxRarity or ItemRarity;
		ItemRarity = (ItemRarity > self.db.profile.MaxAvailRarity) and self.db.profile.MaxAvailRarity or ItemRarity;
		
		if (not sItemLink or not ItemID or not ItemRarity) then return; end
		
		local bonuses, socketBonuses = {}, {};
		local iblInfo = ItemBonusLib:ScanItemLink(sItemLink, false);
		if (iblInfo.bonuses) then
			for k,v in pairs(iblInfo.bonuses) do
				bonuses[k] = v;
			end
		end
		
		-- Add bonuses to consumables (extras/Consumables.lua) as ItemBonusLib doesn't seem to catch it (if it does the select should return records):
		if (AddOn.BonusEP.Consumables and AddOn.BonusEP.Consumables[ItemID] and not bonuses["Consumables Added Allready"] and self.db.profile.Consumables) then
			debug("Adding Consumable bonuses");
			for k,v in pairs(AddOn.BonusEP.Consumables[ItemID]) do
				bonuses[k] = (bonuses[k] or 0) + v;
				bonuses["Consumables Added Allready"] = 0;
			end
		end
		local DataMined = bonuses["Consumables Added Allready"] and AddOn.BonusEP.ConsumablesVersion;
		
		-- Add procs and use bonuses (extras/ProcsAndUse.lua)
		local ProcsAndUse = false;
		if (AddOn.BonusEP.ProcsAndUse and AddOn.BonusEP.ProcsAndUse[ItemID] and not bonuses["ProcsAndUse Added Allready"] and self.db.profile.ProcsAndUse) then
			debug("Adding Procs and Use bonuses");
			for k,v in pairs(AddOn.BonusEP.ProcsAndUse[ItemID]) do
				bonuses[k] = (bonuses[k] or 0) + v;
				bonuses["ProcsAndUse Added Allready"] = 0;
			end
		end
		local ProcsAndUse = bonuses["ProcsAndUse Added Allready"] and AddOn.BonusEP.ProcsAndUseVersion;
		
		-- Add possible inactive gem bonus (always inactive as we strip gems from the link)
		if (iblInfo.inactive_gem_bonus) then
			debug("Adding socket bonuses to socket bonus table");
			for k,v in pairs(iblInfo.inactive_gem_bonus) do
				socketBonuses[k] = v;
			end
		end
		
		-- How much is 1 percent of something worth :)
		local CritPercent = (1 / ItemBonusLib:GetRatingBonus("CR_CRIT", 1));
		local RangedCritPercent = (1 / ItemBonusLib:GetRatingBonus("CR_RANGEDCRIT", 1));
		local ExpertisePercent = (1 / ItemBonusLib:GetRatingBonus("CR_EXPERTISE", 1));
		debug("1%% Crit needs %.2f in Rating for your level", CritPercent);
		debug("1%% Ranged Crit needs %.2f in Rating for your level", RangedCritPercent);
		debug("1 Expertise needs %.2f in Rating for your level", ExpertisePercent);
		
		-- Danger here is ItemSubType are localized on the client!
		local RaceBonus = false; -- Level 70 only as this changes while leveling there's really no point caring much about it at low levels except for possibly twinks
		if (self.db.profile.RaceSpecifics and self.RaceEN == "Dwarf" and ItemSubType == L["Guns"]) then
			-- Dwarf: 1% Crit with guns
			RaceBonus = format(L["Racial bonus [+%.2f] ranged crit rating"], RangedCritPercent);
			bonuses["RANGEDCRIT"] = (bonuses["RANGEDCRIT"] or 0) + RangedCritPercent;
		elseif (self.db.profile.RaceSpecifics and self.RaceEN == "Human" and (ItemSubType == L["One-Handed Maces"] or ItemSubType == L["One-Handed Swords"] or ItemSubType == L["Two-Handed Maces"] or ItemSubType == L["Two-Handed Swords"])) then
			-- Human: 5 Expertise with Sword and Mace
			RaceBonus = format(L["Racial bonus [+%.2f] expertise rating"], (5 * ExpertisePercent));
			bonuses["CR_EXPERTISE"] = (bonuses["CR_EXPERTISE"] or 0) + (5 * ExpertisePercent);
		elseif (self.db.profile.RaceSpecifics and self.RaceEN == "Orc" and (ItemSubType == L["One-Handed Axes"] or ItemSubType == L["Two-Handed Axes"])) then
			-- Orc: 5 Expertise with Axe
			RaceBonus = format(L["Racial bonus [+%.2f] expertise rating"], (5 * ExpertisePercent));
			bonuses["CR_EXPERTISE"] = (bonuses["CR_EXPERTISE"] or 0) + (5 * ExpertisePercent);
		elseif (self.db.profile.RaceSpecifics and self.RaceEN == "Troll" and (ItemSubType == L["Bows"] or ItemSubType == L["Thrown"])) then
			-- Troll: 1% Crit with Throwing Weapon and Bow
			RaceBonus = format(L["Racial bonus [+%.2f] ranged crit rating"], RangedCritPercent);
			bonuses["RANGEDCRIT"] = (bonuses["RANGEDCRIT"] or 0) + RangedCritPercent;
		end
		
		-- Todo: Get full Expertise bonus,  get new Expertise rating with this equipped, take the diff and use that!
		-- Expertise Hack Quick: floor(15/ExpertisePercent)*ExpertisePercent
		local ExpertiseHacked = false;
		if (self.db.profile.HackExpertise and bonuses["CR_EXPERTISE"]) then
			debug("Recalculating Expertise, Old value = [%.2f], New value = [%.2f]", bonuses["CR_EXPERTISE"], floor(bonuses["CR_EXPERTISE"] / ExpertisePercent) * ExpertisePercent);
			bonuses["CR_EXPERTISE"] = floor(bonuses["CR_EXPERTISE"] / ExpertisePercent) * ExpertisePercent;
			ExpertiseHacked = true;
		end
		
		local IgnoreSocketColors, newIgnoreSocketColors, EmptyLineAboveAdded, newEmptyLineAboveAdded = false, false, false, false;
		
		newIgnoreSocketColors, newEmptyLineAboveAdded = self:AddSetLines(tooltip, bonuses, socketBonuses, ItemRarity, EmptyLineAboveAdded);
		IgnoreSocketColors, EmptyLineAboveAdded = (IgnoreSocketColors or newIgnoreSocketColors), (EmptyLineAboveAdded or newEmptyLineAboveAdded);
		
		--[[ Also add saved sets except hidden ones ]]
		for SetName in pairs(self.db.profile.ValueSet) do
			if (not self.db.profile.HiddenSet[SetName]) then
				newIgnoreSocketColors, newEmptyLineAboveAdded = self:AddSetLines(tooltip, bonuses, socketBonuses, ItemRarity, EmptyLineAboveAdded, SetName);
				IgnoreSocketColors, EmptyLineAboveAdded = (IgnoreSocketColors or newIgnoreSocketColors), (EmptyLineAboveAdded or newEmptyLineAboveAdded);
			end
		end
		
		local red, green, blue = AddOn.db.profile.Color.r, AddOn.db.profile.Color.g, AddOn.db.profile.Color.b;
		
		if (IgnoreSocketColors) then
			tooltip:AddLine(ttS .. format(L["%s ignore socket colors"], L["_tooltip Ignore Color"]), AddOn.colors.info.R, AddOn.colors.info.G, AddOn.colors.info.B);
		end
		
		if (RaceBonus) then
			tooltip:AddLine(ttS .. RaceBonus, AddOn.colors.info.R, AddOn.colors.info.G, AddOn.colors.info.B);
		end
		
		if (ExpertiseHacked) then
			tooltip:AddLine(ttS .. format(L["Expertise is modified (math.floor([val]/%.2f)*%.2f)"], ExpertisePercent, ExpertisePercent), AddOn.colors.info.R, AddOn.colors.info.G, AddOn.colors.info.B);
		end
		
		if (ProcsAndUse) then
			ProcsAndUse = ((VersionCompare(ProcsAndUse, AddOn.ClientVersion) == 1) and "|cff" .. rgb2hex(AddOn.colors.critical.R, AddOn.colors.critical.G, AddOn.colors.critical.B) .. ProcsAndUse .. "|r") or ProcsAndUse;
			
			tooltip:AddLine(ttS .. format(L["Proc/use are guestimated (v%s)"], ProcsAndUse), AddOn.colors.info.R, AddOn.colors.info.G, AddOn.colors.info.B);
		end
		
		if (DataMined) then
			DataMined = ((VersionCompare(DataMined, AddOn.ClientVersion) == 1) and "|cff" .. rgb2hex(AddOn.colors.critical.R, AddOn.colors.critical.G, AddOn.colors.critical.B) .. DataMined .. "|r") or DataMined;
			
			tooltip:AddLine(ttS .. format(L["Values are datamined (v%s)"], DataMined), AddOn.colors.info.R, AddOn.colors.info.G, AddOn.colors.info.B);
		end
		
		-- Class specifif things?
		if (type(self[self.ClassEN]) == "function" and self.db.profile.ClassSpecifics) then
			if (self.db.profile.ClassSpecificsColor) then
				red, green, blue = AddOn.colors.class.R, AddOn.colors.class.G, AddOn.colors.class.B;
			end
			
			local classExec, classTable = pcall(self[self.ClassEN], self, sItemLink, ItemID, ItemRarity, bonuses, socketBonuses);
			
			if (classExec) then
				if (#classTable.lefts > 0 and #classTable.lefts >= #classTable.rights) then
					if (classTable.title and classTable.title ~= "") then
						tooltip:AddLine(tostring(classTable.title), red, green, blue);
					end
					
					for i = 1, #classTable.lefts do
						local left = classTable.lefts[i];
						local right = classTable.rights[i];
						
						if (not right or right == "") then
							tooltip:AddLine(ttS .. tostring(left), red, green, blue);
						else
							tooltip:AddDoubleLine(ttS .. tostring(left), tostring(right), red, green, blue, red, green, blue);
						end
					end
				end
			else
				print(L["Error adding class specifics for %s: %s"], self.ClassEN, classTable);
			end
		end
		
		if (EmptyLineAboveAdded and self.db.profile.EmptyLineBelow) then
			tooltip:AddLine(" ", AddOn.db.profile.Color.r, AddOn.db.profile.Color.g, AddOn.db.profile.Color.b);
		end
		
		tooltip:Show();
	end
end

function AddOn:CreateClassTable(title)
	return { lefts = {}, rights = {}, title = title };
end

function AddOn:AddClassLines(tbl, left, right)
	if (not tbl or type(tbl) ~= "table") then tbl = self:CreateClassTable(); end
	tbl.lefts[#tbl.lefts+1] = left or "";
	tbl.rights[#tbl.rights+1] = right or "";
end

function AddOn:SetHasValues(values)
	for key, value in pairs(values) do
		if (value > 0 and key ~= "EMPTY_SOCKET_META") then return true; end
	end
	
	return false;
end

local normalSet = "db.profile"; -- this is the key used for the default set (ie not saved sets); (need it to separate sets for listing gems)
function AddOn:AddSetLines(tooltip, bonuses, socketBonuses, ItemRarity, EmptyLineAboveAdded, SetName)
	if (SetName and not self.db.profile.ValueSet[SetName]) then return false, EmptyLineAboveAdded; end -- Set was not found so ignore it
	
	local calcValues = ((SetName and self.db.profile.ValueSet[SetName]) or self.db.profile.Values);
	local total, kingsTotal, obeyColor, kingsObeyColor, gemsUsed, kingsGemsUsed = 0, 0, true, true, {}, {};
	
	if (self:SetHasValues(calcValues)) then
		gemsUsed, kingsGemsUsed = nil, nil;
		total, kingsTotal, obeyColor, kingsObeyColor, gemsUsed, kingsGemsUsed = self:Calculate(bonuses, socketBonuses, calcValues, ItemRarity);
	end
	
	local IgnoreSocketColors = false;
		
	-- All that calculating in vain :(
	if (not self.db.profile.ShowZero and total == 0 and kingsTotal == 0) then return; end
	if (not self.db.profile.ShowNoBonuses and self:CountTable(bonuses) < 1) then return; end
	
	local red, green, blue = AddOn.db.profile.Color.r, AddOn.db.profile.Color.g, AddOn.db.profile.Color.b;
	
	if (not EmptyLineAboveAdded) then
		if (self.db.profile.EmptyLineAbove) then
			tooltip:AddLine(" ", red, green, blue);
		end
		EmptyLineAboveAdded = true;
	end
	if ((not obeyColor) or (not kingsObeyColor)) then
		IgnoreSocketColors = true;
	end
	
	local SetTitle = (SetName and format(L["Saved [Name (BoK)]"], SetName)) or L["Equivalence (BoK)"];
	
	tooltip:AddDoubleLine(SetTitle, format(L["%.2f%s (%.2f%s)"], total, ((obeyColor and "") or "|cffff0000*|r"), kingsTotal, ((kingsObeyColor and "") or "|cffff0000*|r")), red, green, blue, red, green, blue);
	if ((((#gemsUsed > 0 or #kingsGemsUsed > 0) and self.db.profile.ListGems and not SetName) or ((#gemsUsed > 0 or #kingsGemsUsed > 0) and self.db.profile.ListGems and self.db.profile.ListSetGems and SetName)) and (total > 0 or kingsTotal > 0)) then
		-- Wich one has most gems *sigh*
		loopEnd = #gemsUsed;
		if (#kingsGemsUsed > loopEnd) then loopEnd = #kingsGemsUsed; end
		
		tooltip:AddDoubleLine(" ", L["Gem List:"], red, green, blue, red, green, blue);
		for loopNum=1, loopEnd do
			local gem, kingsGem = gemsUsed[loopNum], kingsGemsUsed[loopNum];
			if (gem == kingsGem) then
				gem = (gem and tostring(gem)) or "";
				kingsGem = "";
			else
				gem = (gem and tostring(gem)) or "";
				kingsGem = (kingsGem and "(".. tostring(kingsGem) .. ")") or "";
			end
			
			tooltip:AddDoubleLine(" ", strtrim(format(L["_tooltip _rightside Gem List"],  gem, kingsGem)), red, green, blue, red, green, blue);
		end
	end
	
	return IgnoreSocketColors, EmptyLineAboveAdded;
end

function AddOn:GetCacheKey(valuesTbl)
	local cacheKey = "";
	for key,value in self:DefaultValues():opairs() do
  	cacheKey = cacheKey .. key .. ((valuesTbl and valuesTbl[key]) or 0) ..";";
	end
	return cacheKey;
end

local statsAffectedByKings = { STR = true, AGI = true, STA = true, INT = true, SPI = true};
local kingsMultiplier = (110 / 100);
function AddOn:Calculate(itemValues, socketValues, calcValues, itemRarity)
	itemValues = itemValues or {};
	socketValues = socketValues or {};
	
	local gemsUsed, kingsGemsUsed = {}, {};
	local GemCacheKeyBase = self:GetCacheKey(calcValues);
	
	local total, kingsTotal, obeyColor, kingsObeyColor = 0, 0, true, true;
	for stat, amount in pairs(itemValues) do
		total = total + ((calcValues[stat] or 0) * amount);
		kingsTotal = kingsTotal + (((calcValues[stat] or 0) * amount) * ((statsAffectedByKings[stat] and kingsMultiplier) or 1));
	end
	
	if (itemValues["EMPTY_SOCKET_META"] or itemValues["EMPTY_SOCKET_RED"] or itemValues["EMPTY_SOCKET_BLUE"] or itemValues["EMPTY_SOCKET_YELLOW"]) then
		
		if (itemValues["EMPTY_SOCKET_META"] and itemValues["EMPTY_SOCKET_META"] > 0) then
		--// META GEMS (can return nil if MaxRarity is below rare
			local gemTotalMeta, kingsGemTotalMeta, _, _, gemNameMeta, kingsGemNameMeta = AddOn:BestGem("META", calcValues, itemRarity, GemCacheKeyBase);
			gemTotalMeta, kingsGemTotalMeta = gemTotalMeta * (itemValues["EMPTY_SOCKET_META"] or 0), kingsGemTotalMeta * (itemValues["EMPTY_SOCKET_META"] or 0);
			
			total = total + gemTotalMeta;
			kingsTotal = kingsTotal + kingsGemTotalMeta;
			
			tinsert(gemsUsed, gemNameMeta);
			tinsert(kingsGemsUsed, kingsGemNameMeta);
		end
		
		local allGems = ((itemValues["EMPTY_SOCKET_RED"] or 0) + (itemValues["EMPTY_SOCKET_BLUE"] or 0) + (itemValues["EMPTY_SOCKET_YELLOW"] or 0));
		if (allGems > 0) then
			local gemTotalAny, kingsGemTotalAny, _, _, gemNameAny, kingsGemNameAny = AddOn:BestGem("ANYCOLOR", calcValues, itemRarity, GemCacheKeyBase);
			gemTotalAny, kingsGemTotalAny = gemTotalAny * allGems, kingsGemTotalAny * allGems;
			
			local gemTotalColor, kingsGemTotalColor = 0, 0;
			local gemTable, kingsGemTable = {}, {};
			for gemType, gemAmount in pairs({ RED = "EMPTY_SOCKET_RED", BLUE = "EMPTY_SOCKET_BLUE", YELLOW = "EMPTY_SOCKET_YELLOW" }) do
				if (itemValues[gemAmount] and itemValues[gemAmount] > 0) then
					local gemColor, kingsGemColor, _, _, gemNameColor, kingsGemNameColor = AddOn:BestGem(gemType, calcValues, itemRarity, GemCacheKeyBase);
					
					if (itemValues[gemAmount] and itemValues[gemAmount] > 0) then
						tinsert(gemTable, gemNameColor);
						tinsert(kingsGemTable, kingsGemNameColor);
					end
					
					gemTotalColor = gemTotalColor + (gemColor * (itemValues[gemAmount] or 0));
					kingsGemTotalColor = kingsGemTotalColor + (kingsGemColor * (itemValues[gemAmount] or 0));
				end
			end
			
			local bonus, kingsBonus = 0, 0;
			for stat, amount in pairs(socketValues) do
				bonus = bonus + ((calcValues[stat] or 0) * amount);
				kingsBonus = kingsBonus + (((calcValues[stat] or 0) * amount) * ((statsAffectedByKings[stat] and kingsMultiplier) or 1));
			end
			
			if (gemTotalAny > (gemTotalColor + bonus)) then
				total = total + gemTotalAny;
				obeyColor = false;
				
				tinsert(gemsUsed, gemNameAny);
			else
				total = total + (gemTotalColor + bonus);
				
				for _, v in pairs(gemTable) do
					tinsert(gemsUsed, v);
				end
			end
			
			if (kingsGemTotalAny > (kingsGemTotalColor + kingsBonus)) then
				kingsTotal = kingsTotal + kingsGemTotalAny;
				kingsObeyColor = false;
				
				tinsert(kingsGemsUsed, kingsGemNameAny);
			else
				kingsTotal = kingsTotal + (kingsGemTotalColor + kingsBonus);
				
				for _, v in pairs(kingsGemTable) do
					tinsert(kingsGemsUsed, v);
				end
			end
		end
	end
	
	return round(total, 2), round(kingsTotal, 2), obeyColor, kingsObeyColor, gemsUsed, kingsGemsUsed;
end

local gemColors = { ["BLUE"] = true, ["YELLOW"] = true, ["RED"] = true, ["META"] = true, ["ANYCOLOR"] = true };
function AddOn:PrintBestGem(color)
	color = (color and gemColors[upper(color)] and upper(color)) or "ANYCOLOR";
	
	local gemValue, kingsGemValue, gemID, kingsGemID, gemName, kingsGemName = AddOn:BestGem(color, self.db.profile.Values, self.db.profile.MaxAvailRarity, self:GetCacheKey(self.db.profile.Values));
	debug(gemID);
	local _, gemLink = gemID and GetItemInfo(gemID);
	debug(kingsGemID);
	local _, kingsGemLink = kingsGemID and GetItemInfo(kingsGemID);
	
	print(L["Best color gem is name at 0"], L[color], (gemLink or gemName), gemValue)
	print(L["Best color gem BoKed is name at 0"], L[color], (kingsGemLink or kingsGemName), kingsGemValue)
end

-- Has to specify META for meta gems, BLUE for blue, RED for red and YELLOW for yellow. Any other value is interpreted as any non meta gem
local gemCache, kingsGemCache = {}, {};
local gemIDCache, kingsGemIDCache, gemNameCache, kingsGemNameCache = {}, {}, {}, {};
function AddOn:BestGem(color, calcValues, itemRarity, GemCacheKeyBase)
	-- set all values to valid values
	color = (color and gemColors[upper(color)] and upper(color)) or "ANYCOLOR";
	itemRarity = tonumber(itemRarity);
	
	local bestValue, kingsBestValue, gemID, gemName, kingsGemID, kingsGemName = 0, 0, nil, L["None"], nil, L["None"];
	
	local cacheKey = GemCacheKeyBase .. ";" .. itemRarity .. ";" .. color;
	if (gemCache[cacheKey] and kingsGemCache[cacheKey]) then
		return gemCache[cacheKey], kingsGemCache[cacheKey], gemIDCache[cacheKey], kingsGemIDCache[cacheKey], gemNameCache[cacheKey], kingsGemNameCache[cacheKey];
	end
	
	if (color == "ANYCOLOR") then
		-- Crap special case :(
		for ID, gemInfo in pairs(AddOn.InfoStorage.Gems) do
			if (not gemInfo["Color"]["META"] and gemInfo["Rarity"] <= itemRarity) then
				local gem, gemKings = self:Calculate(gemInfo["Bonus"], nil, calcValues, itemRarity)
				
				if (gem >= bestValue) then
					bestValue = gem;
					gemID, gemName = ID, gemInfo["Name"];
				end
				
				if (gemKings >= kingsBestValue) then
					kingsBestValue = gemKings;
					kingsGemID, kingsGemName = ID, gemInfo["Name"];
				end
			end
		end
	else
		for ID, gemInfo in pairs(AddOn.InfoStorage.Gems) do
			if (gemInfo["Color"][color] and gemInfo["Rarity"] <= itemRarity) then
				-- Valid gem to calc
				local gem, gemKings = self:Calculate(gemInfo["Bonus"], nil, calcValues, itemRarity)
				
				if (gem >= bestValue) then
					bestValue = gem;
					gemID, gemName = ID, gemInfo["Name"];
				end
				
				if (gemKings >= kingsBestValue) then
					kingsBestValue = gemKings;
					kingsGemID, kingsGemName = ID, gemInfo["Name"];
				end
			end
		end
	end
	
	-- Save in cache
	gemCache[cacheKey], kingsGemCache[cacheKey] = round(bestValue, 2), round(kingsBestValue, 2);
	gemIDCache[cacheKey], kingsGemIDCache[cacheKey], gemNameCache[cacheKey], kingsGemNameCache[cacheKey] = gemID, kingsGemID, gemName, kingsGemName;
	return gemCache[cacheKey], kingsGemCache[cacheKey], gemIDCache[cacheKey], kingsGemIDCache[cacheKey], gemNameCache[cacheKey], kingsGemNameCache[cacheKey];
end

function AddOn:StripInfo(Itemlink)
	local found, _, itemstring = find(Itemlink, "^|c%x+|H(.+)|h%[.+%]")
	if (not found) then return Itemlink; end
	
	local linkType, itemId, enchantId, jewelId1, jewelId2, jewelId3, jewelId4, suffixId, uniqueId = strsplit(":", itemstring)
	enchantId = "0";
	jewelId1 = "0";
	jewelId2 = "0";
	jewelId3 = "0";
	jewelId4 = "0";
	local newItemString = strjoin(":", linkType, itemId, enchantId, jewelId1, jewelId2, jewelId3, jewelId4, suffixId, uniqueId);
	local _, newLink, itemRarity, _, _, ItemType, ItemSubType  = GetItemInfo(newItemString);
	--itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc, itemTexture = GetItemInfo(itemID or "itemString" or "itemName" or "itemLink")
	
	return newLink, tonumber(itemId), tonumber(itemRarity), ItemType, ItemSubType;
end

StaticPopupDialogs["PROWLERPAGECONFIRM"] = { };
function AddOn:RecieveValues(event, text, distribution, sender, ...)
	print(L["Receiving data from %s"], sender);
	
	if (self.db.profile.ReceiveValues) then
		self:Deserialize(text, sender);
	else
		print(L["Import ignored due to settings, you can find the toggle under the import tab in /eqp config"]);
	end
end

function AddOn:SendValues(name)
	-- Sends values Serialized to a named friend :)
	local values = self:Serialize();
	self:SendCommMessage("EQP_VALUES_SENDER", values, "WHISPER", name);
end

function AddOn:Serialize()
	return AceSerializer:Serialize(self:TrimTable(self.db.profile.Values));
end

function AddOn:Deserialize(serialized, SetPrefix)
	local CouldDeserialize, tbl = AceSerializer:Deserialize(serialized);
		
	if (CouldDeserialize and type(tbl) == "table") then
		AddOn:ImportSet(SetPrefix, tbl);
	else
		if (not CouldDeserialize) then
			print(L["Import failed, could not deserialize the data!"]);
		elseif (type(tbl) ~= "table") then
			print(L["Import failed, the data wasn't a table with values!"]);
		end
	end
end

function AddOn:ImportSet(SetPrefix, SetTable, TryWithoutSuffix)
	local SetName = "";
	
	if (TryWithoutSuffix and not self.db.profile.ValueSet[SetPrefix]) then
		SetName = SetPrefix;
	else
		local FreeSpot = 0;
		while (true) do
			FreeSpot = FreeSpot + 1;
			SetName = format(L["Name_001"], SetPrefix, FreeSpot);
			if (not self.db.profile.ValueSet[SetName]) then
				break;
			end
		end
	end
	
	if (not self:VerifyImport(SetTable)) then return; end
	
	-- Clear out unknown keys?
	self.db.profile.ValueSet[SetName] = self:TrimTable(SetTable);
	self.db.profile.HiddenSet[SetName] = true; -- Always hide new sets by default
	print(L["Set imported as [%s], state set to Hidden"], SetName);
end

function AddOn:Import(standardizedData)
	-- Clear all values:
	for key,_ in pairs(self.db.profile.Values) do
		self.db.profile.Values[key] = 0;
	end
	
	if (not self:VerifyImport(standardizedData)) then return; end
	
	local affected = {};
	-- Set new values but don't insert new keys:
	for key, value in pairs(standardizedData) do
		if (self.db.profile.Values[key] and value > 0) then
			self.db.profile.Values[key] = value;
			tinsert(affected, self:TranslateValueKey(key));
		end
	end
	
	print(L["Import complete stats affected:"], unpack(affected));
end

function AddOn:VerifyImport(sourceTbl)
	-- this function should just verify that all data looks fine and return true, otherwise print error message and return false!
	-- It mustn't alter the table in anyway to make it fine!
	
	for key, value in pairs(sourceTbl) do
		if (type(key) ~= "string") then
			print(L["Table is malformed, key %s is not a string"], key);
			return false;
		end
		
		if (type(value) ~= "number") then
			print(L["Table is malformed, value for key %s is not a number (%s)"], key, value);
			return false;
		end
		
		if (self:DefaultValues()[key] == nil) then
			print(L["Table is malformed, key %s is not a valid key"], key);
			return false;
		end
		
		if ((key == "EMPTY_SOCKET_META" and (value < 0 or value > 250)) or (key ~= "EMPTY_SOCKET_META" and (value < 0 or value > 10))) then
			print(L["Table is malformed, value in key %s is too big (%i)"], key, value);
			return false;
		end
	end
	
	if (self:CountTable(sourceTbl) < 1) then
		print(L["Table is empty!"], key, value);
		return false;
	end
	
	return true;
end

-- this trims our EPValue tables
function AddOn:TrimTable(SourceTable)
	for Key in pairs(SourceTable) do
		if (type(SourceTable[Key]) == "number" or tonumber(SourceTable[Key])) then
			SourceTable[Key] = tonumber(SourceTable[Key]);
			if (SourceTable[Key] <= 0) then SourceTable[Key] = nil; end
		else
			debug("Table contains non numeric values [%s = %s]", Key, SourceTable[Key]);
			SourceTable[Key] = nil; -- Crap value here
		end
	end
	
	return copy(SourceTable); -- returns a new copy of the table for usage.
end

function AddOn:ImportPreSets(set)
	if (not set) then
		print(L["Set not specified, import aborted!"]);
		return;
	end
	
	if (self.db.profile.Presets[set]) then
		self:Import(self.db.profile.Presets[set])
		return;
	end
	
	print(L["set [%s] not found, import aborted!"], set);
end

function AddOn:TranslateValueKey(key)
	if (key == "EMPTY_SOCKET_META") then return L["Bonus Value to Meta Gems"]; end -- We don't care what it's called in IBL we use it for meta gem bonus ;)
	
	local ibl = ItemBonusLib:GetBonusFriendlyName(key);
	return (ibl and (ibl == "CR_EXPERTISE") and "Expertise Rating") or ibl or key;
end

function AddOn:CountTable(SourceTable)
	if (not SourceTable) then return 0; end
	
	local Count = 0;
	for k in pairs(SourceTable) do
		Count = Count + 1;
	end
	return Count;
end

function AddOn:SpamSafe(text)
	-- replace ' [dot] ' with '.' and ' [at] ' with '@'
	return gsub(gsub(text, " %[at%] ", "@"), " %[dot%] ", ".");
end